from datetime import datetime
from typing import Annotated
import logging

from authlib.jose import Key
from httpx import AsyncClient
from jwt.exceptions import ExpiredSignatureError, InvalidTokenError
from fastapi import FastAPI, HTTPException, Depends, status
from fastapi.middleware.cors import CORSMiddleware

# from starlette.middleware.sessions import SessionMiddleware
# from authlib.integrations.starlette_client.apps import StarletteOAuth2App
# from authlib.oauth2.rfc6749 import OAuth2Token

from .models import User
from .auth_utils import (
    get_user_from_token,
    UserWithRole,
)
from .settings import settings
from .auth_provider import providers

logger = logging.getLogger("oidc-test")

resource_server = FastAPI()


resource_server.add_middleware(
    CORSMiddleware,
    allow_origins=settings.cors_origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

# SessionMiddleware is required by authlib
# resource_server.add_middleware(
#    SessionMiddleware,
#    secret_key=settings.secret_key,
# )

# Route for OAuth resource server


# Routes for RBAC based tests


@resource_server.get("/public")
async def public() -> dict:
    return {"msg": "Not protected"}


@resource_server.get("/protected")
async def get_protected(user: Annotated[User, Depends(get_user_from_token)]):
    assert user is not None  # Just to keep QA checks happy
    return {"msg": "Only authenticated users can see this"}


@resource_server.get("/protected-by-foorole")
async def get_protected_by_foorole(
    user: Annotated[User, Depends(UserWithRole("foorole"))],
):
    assert user is not None
    return {"msg": "Only users with foorole can see this"}


@resource_server.get("/protected-by-barrole")
async def get_protected_by_barrole(
    user: Annotated[User, Depends(UserWithRole("barrole"))],
):
    assert user is not None
    return {"msg": "Protected by barrole"}


@resource_server.get("/protected-by-foorole-and-barrole")
async def get_protected_by_foorole_and_barrole(
    user: Annotated[User, Depends(UserWithRole("foorole")), Depends(UserWithRole("barrole"))],
):
    assert user is not None  # Just to keep QA checks happy
    return {"msg": "Only users with foorole and barrole can see this"}


@resource_server.get("/protected-by-foorole-or-barrole")
async def get_protected_by_foorole_or_barrole(
    user: Annotated[User, Depends(UserWithRole(["foorole", "barrole"]))],
):
    assert user is not None  # Just to keep QA checks happy
    return {"msg": "Only users with foorole or barrole can see this"}


# @resource_server.get("/introspect")
# async def get_introspect(
#    request: Request,
#    oidc_provider: Annotated[StarletteOAuth2App, Depends(get_oidc_provider)],
#    token: Annotated[OAuth2Token, Depends(get_token)],
# ) -> JSONResponse:
#    assert request is not None  # Just to keep QA checks happy
#    if (url := oidc_provider.server_metadata.get("introspection_endpoint")) is None:
#        raise HTTPException(
#            status_code=status.HTTP_401_UNAUTHORIZED,
#            detail="No introspection endpoint found for the OIDC provider",
#        )
#    if (
#        response := await oidc_provider.post(
#            url,
#            token=token,
#            data={"token": token["access_token"]},
#        )
#    ).is_success:
#        return response.json()
#    else:
#        raise HTTPException(status_code=response.status_code, detail=response.text)


@resource_server.get("/{id}")
async def get_resource_(
    id: str,
    # user: Annotated[User, Depends(get_current_user)],
    # oidc_provider: Annotated[StarletteOAuth2App, Depends(get_oidc_provider)],
    # token: Annotated[OAuth2Token, Depends(get_token)],
    user: Annotated[User, Depends(get_user_from_token)],
):
    """Generic path for testing a resource provided by a provider"""
    return await get_resource(id, user)


async def get_resource(resource_id: str, user: User) -> dict:
    """
    Resource processing: build an informative rely as a simple showcase
    """
    try:
        pname = providers[user.auth_provider_id].name
    except KeyError:
        pname = "?"
    resp = {
        "hello": f"Hi {user.name} from an OAuth resource provider",
        "comment": f"I received a request for '{resource_id}' "
        + f"with an access token signed by {pname}",
    }
    # For the demo, resource resource_id matches a scope get:resource_id,
    # but this has to be refined for production
    required_scope = f"get:{resource_id}"
    # Check if the required scope is in the scopes allowed in userinfo
    try:
        if user.has_scope(required_scope):
            await process(user, resource_id, resp)
        else:
            ## For the showcase, giving a explanation.
            ## Alternatively, raise HTTP_401_UNAUTHORIZED
            raise HTTPException(
                status.HTTP_401_UNAUTHORIZED,
                f"No scope {required_scope} in the access token "
                + "but it is required for accessing this resource",
            )
    except ExpiredSignatureError:
        raise HTTPException(status.HTTP_401_UNAUTHORIZED, "The token's signature has expired")
    except InvalidTokenError:
        raise HTTPException(status.HTTP_401_UNAUTHORIZED, "The token is invalid")
    return resp


async def process(user, resource_id, resp):
    """
    Too simple to be serious.
    It's a good fit for a plugin architecture for production
    """
    assert user is not None
    if resource_id == "time":
        resp["time"] = datetime.now().strftime("%c")
    elif resource_id == "bs":
        async with AsyncClient() as client:
            bs = await client.get("https://corporatebs-generator.sameerkumar.website/")
            resp["bs"] = bs.json().get("phrase", "Sorry, i am out of BS today.")
    else:
        raise HTTPException(
            status.HTTP_401_UNAUTHORIZED, f"I don't known how to give '{resource_id}'."
        )


# assert user.oidc_provider is not None
### Get some info (TODO: refactor)
# if (auth_provider_id := user.oidc_provider.name) is None:
#    raise HTTPException(
#        status.HTTP_401_UNAUTHORIZED,
#        "Request headers must have a 'auth_provider' field",
#    )
# if (
#    auth_provider_settings := oidc_providers_settings.get(auth_provider_id)
# ) is None:
#    raise HTTPException(
#        status.HTTP_401_UNAUTHORIZED, f"Unknown auth provider '{auth_provider_id}'"
#    )
# if (key := auth_provider_settings.get_public_key()) is None:
#    raise HTTPException(
#        status.HTTP_401_UNAUTHORIZED,
#        f"Key for provider '{auth_provider_id}' unknown",
#    )
# logger.warn(f"refresh with scope {scope}")
# breakpoint()
# refreshed_auth_info = await user.oidc_provider.fetch_access_token(scope=scope)
### Decode the new token
# try:
#    payload = decode(
#        refreshed_auth_info["access_token"],
#        key=key,
#        algorithms=["RS256"],
#        audience="account",
#        options={"verify_signature": not settings.insecure.skip_verify_signature},
#    )
# except ExpiredSignatureError as err:
#    logger.info(f"Expired signature: {err}")
#    raise HTTPException(
#        status.HTTP_401_UNAUTHORIZED,
#        "Expired signature (refresh not implemented yet)",
#    )
